
/*    
    This code is under GPL. Please read license terms and conditions at http://www.gnu.org/

    Yann LANGLAIS

    Changelog:
    11/03/2002  Upgrade version to v1.0.0 and make code reentrant w/ pthread_mutexes
*/   

#include <stdlib.h>
#include <strings.h>
#include <string.h>
#include "mem.h"
#ifdef __reentrant__
#include <pthread.h>
#define storage_lock(x)   pthread_mutex_lock((x)->mp)
#define storage_unlock(x) pthread_mutex_unlock((x)->mp)
#else
#define storage_lock(x)
#define storage_unlock(x)
#endif

char storage_NAME[]    = "Simple storage management (dynamic arrays)";
char storage_VERSION[] = "1.0.0";
char storage_DATEVER[] = "11/03/2002";

typedef struct {
	size_t unit;
	size_t grain;
	int    used;
	int    allocated;
	int    pages;
	char **data;
    #ifdef __reentrant__
    pthread_mutex_t  *mp;
    #endif  
} storage_t, *pstorage_t;

#define __storage_cc__
#include "storage.h"
#undef  __storage_cc__

pstorage_t storage_destroy(pstorage_t ps) {
	int i;
	if (ps) { 
		storage_lock(ps);
		/* free all pages: */
		for (i = 0; i < ps->pages; i++) 
			if (ps->data[i]) mem_free(ps->data[i]);
		/* free page array: */
		mem_free(ps->data);
		storage_unlock(ps);
		#ifdef __reentrant__
		pthread_mutex_destroy(ps->mp);
		#endif
		
		mem_free(ps);
	}
	return NULL;
}

size_t storage_sizeof() { return sizeof(storage_t); }

pstorage_t storage_new(size_t unit, size_t grain) {
	pstorage_t ps;
	if (!(ps = (pstorage_t) mem_zmalloc(sizeof(storage_t)))) return NULL;
	if (!(ps->data = (char **) mem_zmalloc(sizeof(char *)))) return storage_destroy(ps);
	if (!(ps->data[0] = (char *) mem_zmalloc(unit * grain))) return storage_destroy(ps);
	ps->grain = grain;
	ps->unit  = unit;
	ps->used  = 0;
	ps->pages = 1;
	ps->allocated = grain;
	#ifdef __reentrant__
	ps->mp = (pthread_mutex_t *) mem_zmalloc(sizeof(pthread_mutex_t));
	#endif
	return ps;
}

char *storage_add(pstorage_t ps, char *pdata) {
	char *p;
	if (!ps) return NULL;
	storage_lock(ps);	
	if (ps->used >= ps->allocated - 1) {
		/* Need more room! 
		 * Room extension CANNOT be done by resizing since realloc may change storage location.
		 * Thus, in this case,  pointers refering to stored data within the storage area
		 * would point to anywhere but expected data!!!
		 * Instead, room extension can be done by resizing a page array and allocate a 
		 * new page of grain * unit bytes. This way, we are sure that allocated areas will
		 * not be moved.
		 */
		
		/* resize page array: */
		if (!(p = (char *) mem_realloc((char *) ps->data, (ps->pages + 1) * sizeof(char *)))) {
			storage_unlock(ps);
			return NULL;
		}
		ps->pages++;
		ps->data = (char **) p;

		/* allocate a new page at the end of the page array: */
		if (!(ps->data[ps->pages - 1] = (char *) mem_zmalloc(ps->grain * ps->unit))) {
			storage_unlock(ps);
			return NULL;
		}
		ps->allocated += ps->grain;
	}
	/* retrieve page: */
	p = ps->data[ps->used / ps->grain];

	/*retrieve offset with the page: */
	p += (ps->used % ps->grain) * ps->unit; 
	/* p = (ps->data + (ps->used * ps->unit)); */
	memcpy((void *) p, (void *) pdata, ps->unit);
	ps->used++;
	storage_unlock(ps);
	return p;
}
		
char * storage_get(pstorage_t ps, int i) {
	if (!ps) return NULL;
	if (i < 0 || i > ps->used) return NULL;
	return ps->data[i / ps->grain] + ((i % ps->grain) * ps->unit);
}

int storage_used(pstorage_t ps) {
	if (ps) return ps->used;
	return -1;
}

int storage_allocated(pstorage_t ps) {
	if (ps) return ps->allocated;
	return -1;
}

int storage_available(pstorage_t ps) {
	if (ps) return ps->allocated - ps->used;
	return -1;
}

int storage_grain(pstorage_t ps) {
	if (ps) return ps->grain;
	return -1;
}

int storage_unit(pstorage_t ps) {
	if (ps) return ps->unit;
	return -1;
}

int storage_pages(pstorage_t ps) {
	if (ps) return ps->pages;
	return -1;
}
